id ICON "shp2pgsql-gui.ico"
1 VERSIONINFO
FILEVERSION     1,0,0,0
PRODUCTVERSION  1,0,0,0
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904E4"
    BEGIN
      VALUE "CompanyName", "Open Source Geospatial Consortium"
      VALUE "FileDescription", "Shape to PostGIS Converter"
      VALUE "FileVersion", "1.0"
      VALUE "InternalName", "shp2pgsql-gui"
      VALUE "LegalCopyright", "Copyright (c) OSGeo"
      VALUE "OriginalFilename", "shp2pgsql-gui.exe"
      VALUE "ProductName", "Shp2PgSQL-GUI"
      VALUE "ProductVersion", "1.0"
    END
  END

  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x409, 1252
  END
END
