SELECT 'GEOMETRY(0)', postgis_type_name('GEOMETRY', 0);
SELECT 'GEOMETRY(1)', postgis_type_name('GEOMETRY', 1);
SELECT 'GEOMETRY(2)', postgis_type_name('GEOMETRY', 2);
SELECT 'GEOMETRY(3)', postgis_type_name('GEOMETRY', 3);
SELECT 'GEOMETRY(4)', postgis_type_name('GEOMETRY', 4);
SELECT 'GEOMETRY(5)', postgis_type_name('GEOMETRY', 5);
SELECT 'GEOMETRYM(2)', postgis_type_name('GEOMETRYM', 2);
SELECT 'GEOMETRYM(3)', postgis_type_name('GEOMETRYM', 3);
SELECT 'GEOMETRYZ(3)', postgis_type_name('GEOMETRYZ', 3);
SELECT 'GEOMETRYZM(3)', postgis_type_name('GEOMETRYZM', 3);
SELECT 'GEOMETRYZM(4)', postgis_type_name('GEOMETRYZM', 4);

SELECT 'POINT(0)', postgis_type_name('POINT', 0);
SELECT 'POINT(1)', postgis_type_name('POINT', 1);
SELECT 'POINT(2)', postgis_type_name('POINT', 2);
SELECT 'POINT(3)', postgis_type_name('POINT', 3);
SELECT 'POINT(4)', postgis_type_name('POINT', 4);
SELECT 'POINT(5)', postgis_type_name('POINT', 5);
SELECT 'POINTM(2)', postgis_type_name('POINTM', 2);
SELECT 'POINTM(3)', postgis_type_name('POINTM', 3);
SELECT 'POINTZ(3)', postgis_type_name('POINTZ', 3);
SELECT 'POINTZM(3)', postgis_type_name('POINTZM', 3);
SELECT 'POINTZM(4)', postgis_type_name('POINTZM', 4);

SELECT 'LINESTRING(0)', postgis_type_name('LINESTRING', 0);
SELECT 'LINESTRING(1)', postgis_type_name('LINESTRING', 1);
SELECT 'LINESTRING(2)', postgis_type_name('LINESTRING', 2);
SELECT 'LINESTRING(3)', postgis_type_name('LINESTRING', 3);
SELECT 'LINESTRING(4)', postgis_type_name('LINESTRING', 4);
SELECT 'LINESTRING(5)', postgis_type_name('LINESTRING', 5);
SELECT 'LINESTRINGM(2)', postgis_type_name('LINESTRINGM', 2);
SELECT 'LINESTRINGM(3)', postgis_type_name('LINESTRINGM', 3);
SELECT 'LINESTRINGZ(3)', postgis_type_name('LINESTRINGZ', 3);
SELECT 'LINESTRINGZM(3)', postgis_type_name('LINESTRINGZM', 3);
SELECT 'LINESTRINGZM(4)', postgis_type_name('LINESTRINGZM', 4);

SELECT 'POLYGON(0)', postgis_type_name('POLYGON', 0);
SELECT 'POLYGON(1)', postgis_type_name('POLYGON', 1);
SELECT 'POLYGON(2)', postgis_type_name('POLYGON', 2);
SELECT 'POLYGON(3)', postgis_type_name('POLYGON', 3);
SELECT 'POLYGON(4)', postgis_type_name('POLYGON', 4);
SELECT 'POLYGON(5)', postgis_type_name('POLYGON', 5);
SELECT 'POLYGONM(2)', postgis_type_name('POLYGONM', 2);
SELECT 'POLYGONM(3)', postgis_type_name('POLYGONM', 3);
SELECT 'POLYGONZ(3)', postgis_type_name('POLYGONZ', 3);
SELECT 'POLYGONZM(3)', postgis_type_name('POLYGONZM', 3);
SELECT 'POLYGONZM(4)', postgis_type_name('POLYGONZM', 4);

SELECT 'MULTIPOINT(0)', postgis_type_name('MULTIPOINT', 0);
SELECT 'MULTIPOINT(1)', postgis_type_name('MULTIPOINT', 1);
SELECT 'MULTIPOINT(2)', postgis_type_name('MULTIPOINT', 2);
SELECT 'MULTIPOINT(3)', postgis_type_name('MULTIPOINT', 3);
SELECT 'MULTIPOINT(4)', postgis_type_name('MULTIPOINT', 4);
SELECT 'MULTIPOINT(5)', postgis_type_name('MULTIPOINT', 5);
SELECT 'MULTIPOINTM(2)', postgis_type_name('MULTIPOINTM', 2);
SELECT 'MULTIPOINTM(3)', postgis_type_name('MULTIPOINTM', 3);
SELECT 'MULTIPOINTZ(3)', postgis_type_name('MULTIPOINTZ', 3);
SELECT 'MULTIPOINTZM(3)', postgis_type_name('MULTIPOINTZM', 3);
SELECT 'MULTIPOINTZM(4)', postgis_type_name('MULTIPOINTZM', 4);

SELECT 'MULTILINESTRING(0)', postgis_type_name('MULTILINESTRING', 0);
SELECT 'MULTILINESTRING(1)', postgis_type_name('MULTILINESTRING', 1);
SELECT 'MULTILINESTRING(2)', postgis_type_name('MULTILINESTRING', 2);
SELECT 'MULTILINESTRING(3)', postgis_type_name('MULTILINESTRING', 3);
SELECT 'MULTILINESTRING(4)', postgis_type_name('MULTILINESTRING', 4);
SELECT 'MULTILINESTRING(5)', postgis_type_name('MULTILINESTRING', 5);
SELECT 'MULTILINESTRINGM(2)', postgis_type_name('MULTILINESTRINGM', 2);
SELECT 'MULTILINESTRINGM(3)', postgis_type_name('MULTILINESTRINGM', 3);
SELECT 'MULTILINESTRINGZ(3)', postgis_type_name('MULTILINESTRINGZ', 3);
SELECT 'MULTILINESTRINGZM(3)', postgis_type_name('MULTILINESTRINGZM', 3);
SELECT 'MULTILINESTRINGZM(4)', postgis_type_name('MULTILINESTRINGZM', 4);

SELECT 'MULTIPOLYGON(0)', postgis_type_name('MULTIPOLYGON', 0);
SELECT 'MULTIPOLYGON(1)', postgis_type_name('MULTIPOLYGON', 1);
SELECT 'MULTIPOLYGON(2)', postgis_type_name('MULTIPOLYGON', 2);
SELECT 'MULTIPOLYGON(3)', postgis_type_name('MULTIPOLYGON', 3);
SELECT 'MULTIPOLYGON(4)', postgis_type_name('MULTIPOLYGON', 4);
SELECT 'MULTIPOLYGON(5)', postgis_type_name('MULTIPOLYGON', 5);
SELECT 'MULTIPOLYGONM(2)', postgis_type_name('MULTIPOLYGONM', 2);
SELECT 'MULTIPOLYGONM(3)', postgis_type_name('MULTIPOLYGONM', 3);
SELECT 'MULTIPOLYGONZ(3)', postgis_type_name('MULTIPOLYGONZ', 3);
SELECT 'MULTIPOLYGONZM(3)', postgis_type_name('MULTIPOLYGONZM', 3);
SELECT 'MULTIPOLYGONZM(4)', postgis_type_name('MULTIPOLYGONZM', 4);

SELECT 'GEOMETRYCOLLECTION(0)', postgis_type_name('GEOMETRYCOLLECTION', 0);
SELECT 'GEOMETRYCOLLECTION(1)', postgis_type_name('GEOMETRYCOLLECTION', 1);
SELECT 'GEOMETRYCOLLECTION(2)', postgis_type_name('GEOMETRYCOLLECTION', 2);
SELECT 'GEOMETRYCOLLECTION(3)', postgis_type_name('GEOMETRYCOLLECTION', 3);
SELECT 'GEOMETRYCOLLECTION(4)', postgis_type_name('GEOMETRYCOLLECTION', 4);
SELECT 'GEOMETRYCOLLECTION(5)', postgis_type_name('GEOMETRYCOLLECTION', 5);
SELECT 'GEOMETRYCOLLECTIONM(2)', postgis_type_name('GEOMETRYCOLLECTIONM', 2);
SELECT 'GEOMETRYCOLLECTIONM(3)', postgis_type_name('GEOMETRYCOLLECTIONM', 3);
SELECT 'GEOMETRYCOLLECTIONZ(3)', postgis_type_name('GEOMETRYCOLLECTIONZ', 3);
SELECT 'GEOMETRYCOLLECTIONZM(3)', postgis_type_name('GEOMETRYCOLLECTIONZM', 3);
SELECT 'GEOMETRYCOLLECTIONZM(4)', postgis_type_name('GEOMETRYCOLLECTIONZM', 4);

SELECT 'CIRCULARSTRING(0)', postgis_type_name('CIRCULARSTRING', 0);
SELECT 'CIRCULARSTRING(1)', postgis_type_name('CIRCULARSTRING', 1);
SELECT 'CIRCULARSTRING(2)', postgis_type_name('CIRCULARSTRING', 2);
SELECT 'CIRCULARSTRING(3)', postgis_type_name('CIRCULARSTRING', 3);
SELECT 'CIRCULARSTRING(4)', postgis_type_name('CIRCULARSTRING', 4);
SELECT 'CIRCULARSTRING(5)', postgis_type_name('CIRCULARSTRING', 5);
SELECT 'CIRCULARSTRINGM(2)', postgis_type_name('CIRCULARSTRINGM', 2);
SELECT 'CIRCULARSTRINGM(3)', postgis_type_name('CIRCULARSTRINGM', 3);
SELECT 'CIRCULARSTRINGZ(3)', postgis_type_name('CIRCULARSTRINGZ', 3);
SELECT 'CIRCULARSTRINGZM(3)', postgis_type_name('CIRCULARSTRINGZM', 3);
SELECT 'CIRCULARSTRINGZM(4)', postgis_type_name('CIRCULARSTRINGZM', 4);

SELECT 'COMPOUNDCURVE(0)', postgis_type_name('COMPOUNDCURVE', 0);
SELECT 'COMPOUNDCURVE(1)', postgis_type_name('COMPOUNDCURVE', 1);
SELECT 'COMPOUNDCURVE(2)', postgis_type_name('COMPOUNDCURVE', 2);
SELECT 'COMPOUNDCURVE(3)', postgis_type_name('COMPOUNDCURVE', 3);
SELECT 'COMPOUNDCURVE(4)', postgis_type_name('COMPOUNDCURVE', 4);
SELECT 'COMPOUNDCURVE(5)', postgis_type_name('COMPOUNDCURVE', 5);
SELECT 'COMPOUNDCURVEM(2)', postgis_type_name('COMPOUNDCURVEM', 2);
SELECT 'COMPOUNDCURVEM(3)', postgis_type_name('COMPOUNDCURVEM', 3);
SELECT 'COMPOUNDCURVEZ(3)', postgis_type_name('COMPOUNDCURVEZ', 3);
SELECT 'COMPOUNDCURVEZM(3)', postgis_type_name('COMPOUNDCURVEZM', 3);
SELECT 'COMPOUNDCURVEZM(4)', postgis_type_name('COMPOUNDCURVEZM', 4);

SELECT 'CURVEPOLYGON(0)', postgis_type_name('CURVEPOLYGON', 0);
SELECT 'CURVEPOLYGON(1)', postgis_type_name('CURVEPOLYGON', 1);
SELECT 'CURVEPOLYGON(2)', postgis_type_name('CURVEPOLYGON', 2);
SELECT 'CURVEPOLYGON(3)', postgis_type_name('CURVEPOLYGON', 3);
SELECT 'CURVEPOLYGON(4)', postgis_type_name('CURVEPOLYGON', 4);
SELECT 'CURVEPOLYGON(5)', postgis_type_name('CURVEPOLYGON', 5);
SELECT 'CURVEPOLYGONM(2)', postgis_type_name('CURVEPOLYGONM', 2);
SELECT 'CURVEPOLYGONM(3)', postgis_type_name('CURVEPOLYGONM', 3);
SELECT 'CURVEPOLYGONZ(3)', postgis_type_name('CURVEPOLYGONZ', 3);
SELECT 'CURVEPOLYGONZM(3)', postgis_type_name('CURVEPOLYGONZM', 3);
SELECT 'CURVEPOLYGONZM(4)', postgis_type_name('CURVEPOLYGONZM', 4);

SELECT 'MULTICURVE(0)', postgis_type_name('MULTICURVE', 0);
SELECT 'MULTICURVE(1)', postgis_type_name('MULTICURVE', 1);
SELECT 'MULTICURVE(2)', postgis_type_name('MULTICURVE', 2);
SELECT 'MULTICURVE(3)', postgis_type_name('MULTICURVE', 3);
SELECT 'MULTICURVE(4)', postgis_type_name('MULTICURVE', 4);
SELECT 'MULTICURVE(5)', postgis_type_name('MULTICURVE', 5);
SELECT 'MULTICURVEM(2)', postgis_type_name('MULTICURVEM', 2);
SELECT 'MULTICURVEM(3)', postgis_type_name('MULTICURVEM', 3);
SELECT 'MULTICURVEZ(3)', postgis_type_name('MULTICURVEZ', 3);
SELECT 'MULTICURVEZM(3)', postgis_type_name('MULTICURVEZM', 3);
SELECT 'MULTICURVEZM(4)', postgis_type_name('MULTICURVEZM', 4);

SELECT 'MULTISURFACE(0)', postgis_type_name('MULTISURFACE', 0);
SELECT 'MULTISURFACE(1)', postgis_type_name('MULTISURFACE', 1);
SELECT 'MULTISURFACE(2)', postgis_type_name('MULTISURFACE', 2);
SELECT 'MULTISURFACE(3)', postgis_type_name('MULTISURFACE', 3);
SELECT 'MULTISURFACE(4)', postgis_type_name('MULTISURFACE', 4);
SELECT 'MULTISURFACE(5)', postgis_type_name('MULTISURFACE', 5);
SELECT 'MULTISURFACEM(2)', postgis_type_name('MULTISURFACEM', 2);
SELECT 'MULTISURFACEM(3)', postgis_type_name('MULTISURFACEM', 3);
SELECT 'MULTISURFACEZ(3)', postgis_type_name('MULTISURFACEZ', 3);
SELECT 'MULTISURFACEZM(3)', postgis_type_name('MULTISURFACEZM', 3);
SELECT 'MULTISURFACEZM(4)', postgis_type_name('MULTISURFACEZM', 4);

SELECT 'POLYHEDRALSURFACE(0)', postgis_type_name('POLYHEDRALSURFACE', 0);
SELECT 'POLYHEDRALSURFACE(1)', postgis_type_name('POLYHEDRALSURFACE', 1);
SELECT 'POLYHEDRALSURFACE(2)', postgis_type_name('POLYHEDRALSURFACE', 2);
SELECT 'POLYHEDRALSURFACE(3)', postgis_type_name('POLYHEDRALSURFACE', 3);
SELECT 'POLYHEDRALSURFACE(4)', postgis_type_name('POLYHEDRALSURFACE', 4);
SELECT 'POLYHEDRALSURFACE(5)', postgis_type_name('POLYHEDRALSURFACE', 5);
SELECT 'POLYHEDRALSURFACEM(2)', postgis_type_name('POLYHEDRALSURFACEM', 2);
SELECT 'POLYHEDRALSURFACEM(3)', postgis_type_name('POLYHEDRALSURFACEM', 3);
SELECT 'POLYHEDRALSURFACEZ(3)', postgis_type_name('POLYHEDRALSURFACEZ', 3);
SELECT 'POLYHEDRALSURFACEZM(3)', postgis_type_name('POLYHEDRALSURFACEZM', 3);
SELECT 'POLYHEDRALSURFACEZM(4)', postgis_type_name('POLYHEDRALSURFACEZM', 4);

SELECT 'TRIANGLE(0)', postgis_type_name('TRIANGLE', 0);
SELECT 'TRIANGLE(1)', postgis_type_name('TRIANGLE', 1);
SELECT 'TRIANGLE(2)', postgis_type_name('TRIANGLE', 2);
SELECT 'TRIANGLE(3)', postgis_type_name('TRIANGLE', 3);
SELECT 'TRIANGLE(4)', postgis_type_name('TRIANGLE', 4);
SELECT 'TRIANGLE(5)', postgis_type_name('TRIANGLE', 5);
SELECT 'TRIANGLEM(2)', postgis_type_name('TRIANGLEM', 2);
SELECT 'TRIANGLEM(3)', postgis_type_name('TRIANGLEM', 3);
SELECT 'TRIANGLEZ(3)', postgis_type_name('TRIANGLEZ', 3);
SELECT 'TRIANGLEZM(3)', postgis_type_name('TRIANGLEZM', 3);
SELECT 'TRIANGLEZM(4)', postgis_type_name('TRIANGLEZM', 4);

SELECT 'TIN(0)', postgis_type_name('TIN', 0);
SELECT 'TIN(1)', postgis_type_name('TIN', 1);
SELECT 'TIN(2)', postgis_type_name('TIN', 2);
SELECT 'TIN(3)', postgis_type_name('TIN', 3);
SELECT 'TIN(4)', postgis_type_name('TIN', 4);
SELECT 'TIN(5)', postgis_type_name('TIN', 5);
SELECT 'TINM(2)', postgis_type_name('TINM', 2);
SELECT 'TINM(3)', postgis_type_name('TINM', 3);
SELECT 'TINZ(3)', postgis_type_name('TINZ', 3);
SELECT 'TINZM(3)', postgis_type_name('TINZM', 3);
SELECT 'TINZM(4)', postgis_type_name('TINZM', 4);

