SELECT ST_Scroll('POINT(0 0)', 'POINT(0 0)');
SELECT ST_Scroll('LINESTRING(0 0, 10 0, 5 5,0 0)', 'LINESTRING(0 0,1 0)');
SELECT ST_Scroll('LINESTRING(0 0, 10 0, 5 5)', 'POINT(0 0)');

SELECT ST_AsEWKT(ST_Scroll('SRID=4326;LINESTRING(0 0, 10 0, 5 5,0 0)', 'POINT(0 0)'));
SELECT ST_AsEWKT(ST_Scroll('SRID=4326;LINESTRING(0 0, 10 0, 5 5,0 0)', 'POINT(5 5)'));
SELECT ST_AsEWKT(ST_Scroll('SRID=4326;LINESTRING(0 0, 10 0, 5 5,0 0)', 'POINT(10 0)'));

SELECT ST_AsEWKT(ST_Scroll('SRID=4326;LINESTRINGM(0 0 0, 10 0 2, 5 5 4,0 0 0)', 'POINT(10 0 2)'));

SELECT ST_AsEWKT(ST_Scroll('SRID=4326;LINESTRINGZ(0 0 0, 10 0 2, 5 5 4,0 0 0)', 'POINT(5 5 4)'));

SELECT ST_AsEWKT(ST_Scroll('SRID=4326;LINESTRING(0 0 0 1, 10 0 2 0, 5 5 4 2,0 0 0 1)', 'POINT(5 5 4 2)'));
